import logging
import os
import pickle
import subprocess
from datetime import datetime

from git import GitDB, Repo  # pyright: ignore [reportPrivateImportUsage]

from .actions import *
from .config import LennyBotConfig
from .model import LennyBotState
from .model.plan import LennyBotPlan
from .service import ApplyService, GitHubService, PlanService


class LennyBot:
    def __init__(self, config_path):
        self._config = LennyBotConfig(config_path)
        self._log = logging.getLogger(self.__class__.__name__)
        self._github_service = GitHubService(self._config)
        self._plan_service = PlanService(self._github_service, self._config)
        self._apply_service = ApplyService()
        self._repo = None
        self._branch_name = None

    def plan(self) -> LennyBotPlan:
        self._log.info("Start planning...")
        state = LennyBotState(self._config)
        plan = self._plan_service.plan(state)
        self._log.info(f"Created plan with {len(plan.actions)} actions")
        return plan

    def apply(self, plan: LennyBotPlan):
        self._log.info("Start apply...")
        result = self._apply_service.apply(plan)
        self._log.info(f"Applied {len(plan.actions)} actions")
        return result

    def load_plan(self, filename: str) -> LennyBotPlan:
        self._log.debug(f"Load plan from '{filename}'")
        with open(filename, "rb") as file_ptr:
            return pickle.load(file_ptr)

    def save_plan(self, filename: str, plan: LennyBotPlan):
        self._log.debug(f"Store plan to '{filename}'")
        with open(filename, "wb") as file_ptr:
            pickle.dump(plan, file_ptr)

    def ci_setup(self):
        self._log.debug(f"Setup CI")
        now = datetime.now().strftime("%Y%m%d%H%M%S")
        self._branch_name = f"{self._config.github_pr.branch_prefix}"
        if not self._branch_name.endswith("-"):
            self._branch_name = f"{self._branch_name}-"
        self._branch_name = f"{self._branch_name}{now}"
        self._log.debug(f"Determined branch name {self._branch_name}")
        result = subprocess.call(["git", "config", "--global", "--add", "safe.directory", os.getcwd()])
        if result != 0:
            self._log.error("Unexpected return code from git config")
        self._repo = Repo("./", odbt=GitDB)  # type: ignore
        self._log.debug(f"Initialized repository")
        head = self._repo.create_head(self._branch_name)
        self._log.debug(f"Created Head")
        head.checkout()
        self._log.info(f"Working branch is {self._branch_name}")

    def ci_finalize(self, plan: LennyBotPlan, result):
        self._log.debug(f"Finalize CI")
        if self._repo is None:
            raise Exception("Repository is non, ci_setup was not called")
        if not self._repo.index.diff(None) and not self._repo.untracked_files:
            return
        title = f"Lennybot updated {len(plan.applications)} applications"
        if len(plan.applications) == 1:
            title = f"Lennybot updated {plan.applications[0]}"
        body = "Lennybot updated following applications:\n"
        for app in plan.applications:
            body += f"* Bumps {app} from {plan.source_version(app)} to {plan.target_version(app)}\n"
        body += "\n"
        body += "To resolve any conflict try to run the lennybot again\n"
        # Git Commit and Push
        self._repo.config_writer().set_value("user", "name", "lennybot").release()
        self._repo.config_writer().set_value("user", "email", "lennybot@raynigon.com").release()
        self._repo.git.add(A=True)
        self._repo.git.commit(m=title)
        self._repo.git.push("--set-upstream", "origin", self._branch_name)
        # Create Pull Request
        self._github_service.create_pr(self._branch_name, title, body)



from types import SimpleNamespace

from yamlpath import Processor
from yamlpath.common import Parsers
from yamlpath.wrappers import ConsolePrinter

from ..config.config import LennyBotActionConfig
from .iaction import IAction


class UpdateYamlAction(IAction):
    def __init__(self, name, source_version, target_version, config: LennyBotActionConfig) -> None:
        self._name = name
        self._source_version = source_version
        self._target_version = target_version
        if config.target_file is None:
            raise Exception("Target file is not set for application " + name)
        self._target_file = config.target_file
        if config.yaml_path is None:
            raise Exception("YAML Path is not set for application " + name)
        self._yaml_path = config.yaml_path
        if config.value_pattern is not None:
            self._value_pattern = config.value_pattern
        else:
            self._value_pattern = "{{version}}"

    @property
    def application(self) -> str:
        return self._name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def run(self):
        logging_args = SimpleNamespace(quiet=True, verbose=False, debug=False)
        log = ConsolePrinter(logging_args)
        yaml = Parsers.get_yaml_editor()
        (yaml_data, doc_loaded) = Parsers.get_yaml_data(yaml, log, self._target_file)
        if not doc_loaded:
            raise Exception("Yaml document could not be loaded")
        processor = Processor(log, yaml_data)
        processor.set_value(self._yaml_path, self._create_value())
        with open(self._target_file, "w", encoding="utf-8") as file_ptr:
            yaml.dump(yaml_data, stream=file_ptr)

    def _create_value(self):
        return self._value_pattern.replace("{{version}}", self._target_version)



import logging
from typing import List

from lennybot.check import create_check

from ..actions import IAction, create_action
from ..config import LennyBotAppConfig, LennyBotConfig
from ..helper import semver_2_vc
from ..model import LennyBotPlan, LennyBotState
from .github import GitHubService
from .source import create_source


class LennyBotApplication:
    def __init__(self, config: LennyBotAppConfig, global_config: LennyBotConfig, github) -> None:
        self._log = logging.getLogger(self.__class__.__name__)
        self._name = config.name
        self._config = config
        self._global_config = global_config
        self._source = create_source(self._name, config.source, github)
        self._checks = []
        self._action_configs = config.actions
        self._current_version = None
        self._latest_version = None

    @property
    def name(self) -> str:
        return self._name

    def init(self, state: LennyBotState):
        self._current_version = state.current_version(self._name)
        self._latest_version = self._source.latest_version()
        for config in self._config._checks:
            check = create_check(self.name, self._current_version, self._latest_version, config, self._global_config)
            self._checks.append(check)

    def should_update(self) -> bool:
        if self._latest_version is None:
            raise Exception("Application is initialized")
        if self._current_version == self._latest_version:
            return False
        current_vc = semver_2_vc(self._current_version)
        latest_vc = semver_2_vc(self._latest_version)

        if current_vc >= latest_vc:
            return False

        for check in self._checks:
            if not check.check():
                self._log.info("Check '%s' failed for application '%s'", check.__class__.__name__, self.name)
                return False
        return True

    def create_actions(self) -> List[IAction]:
        if self._latest_version is None:
            raise Exception("Application is initialized")
        result = []
        for config in self._action_configs:
            action = create_action(self.name, self._current_version, self._latest_version, config)
            result.append(action)
        return result


class PlanService:
    def __init__(self, github: GitHubService, config: LennyBotConfig) -> None:
        self._log = logging.getLogger(self.__class__.__name__)
        self._github = github
        self._applications: List[LennyBotApplication] = []
        for app_config in config.applications:
            self._applications.append(LennyBotApplication(app_config, config, self._github))

    def plan(self, state: LennyBotState) -> LennyBotPlan:
        actions = []
        for app in self._applications:
            try:
                app.init(state)
                if app.should_update():
                    actions.extend(app.create_actions())
            except Exception as exception:
                self._log.error(f"Exception during action planning for {app.name}")
                raise exception
        return LennyBotPlan(state, actions)



import argparse
import logging
import os

from .lennybot import LennyBot


def _version() -> str:
    if "__version__" in locals():
        return locals()["__version__"]
    if "__version__" in globals():
        return globals()["__version__"]
    if os.path.exists("version.txt"):
        with open("version.txt", encoding="utf-8") as file_ptr:
            return file_ptr.read()
    return "-"


def _find_config(args):
    if args.config is not None:
        for item in args.config:
            return item
    if "LB_CONFIG_FILE" in os.environ.keys():
        return os.environ["LB_CONFIG_FILE"]
    if os.path.exists("config.yaml"):
        return "config.yaml"
    if os.path.exists("config.yml"):
        return "config.yml"
    raise Exception(
        "Configuration file not found, create 'config.yaml' or set LB_CONFIG_FILE to point to the config file"
    )


def _arg_parser():
    parser = argparse.ArgumentParser(
        prog="lennybot",
        description="""
            The Lennybot checks for updates, creates a plan to update resources and can also apply this plan.
        """,
    )
    parser.add_argument(
        "action",
        metavar="action",
        type=str,
        nargs="?",
        choices=["ci", "plan", "apply"],
        default="ci",
        help='The action which should be executed. Has to be one of "plan", "apply" or "ci". Default is "ci"',
    )
    parser.add_argument(
        "-p",
        "--plan",
        dest="plan",
        type=str,
        required=False,
        help="The filename of plan which should be saved or loaded",
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config",
        type=str,
        required=False,
        action="append",
        help="A config value in the format key.subkey=value",
    )
    parser.add_argument("-v", "--version", action="version", version=_version())
    return parser.parse_args()


def main() -> int:
    args = _arg_parser()
    config_file = _find_config(args)
    app = LennyBot(config_file)

    # Execute plan only
    if args.action == "plan":
        plan = app.plan()
        app.save_plan("lennybot.plan", plan)
    # Execute plan and apply
    elif args.action == "apply" and args.plan is None:
        plan = app.plan()
        app.apply(plan)
    # Execute apply with given plan
    elif args.action == "apply" and args.plan is not None:
        plan = app.load_plan(args.plan)
        app.apply(plan)
    elif args.action == "ci":
        app.ci_setup()
        plan = app.plan()
        result = app.apply(plan)
        app.ci_finalize(plan, result)
    else:
        logging.error("Unexpected Arguments", args)
        return 1
    return 0



import yaml

from ..config.config import LennyBotActionConfig
from .iaction import IAction


class UpdateImageTagAction(IAction):
    def __init__(self, name, source_version, target_version, config: LennyBotActionConfig) -> None:
        self._name = name
        self._image = config.image
        if config.kustomize_path is None:
            raise Exception("Kustomize Path file is not set for application " + name)
        self._kustomize_path = config.kustomize_path
        self._source_version = source_version
        self._target_version = target_version
        if config.tag_pattern is not None:
            self._tag_pattern = config.tag_pattern
        else:
            self._tag_pattern = "{{version}}"

    @property
    def application(self) -> str:
        return self._name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def run(self):
        with open(self._kustomize_path, encoding="utf-8") as file_ptr:
            kustomize = yaml.safe_load(file_ptr)
        found = False
        for image in kustomize["images"]:
            if image["name"] != self._image:
                continue
            image["newTag"] = self._create_new_tag()
            found = True
            break
        if not found:
            raise Exception(f"Unable to find image {self._image}")
        with open(self._kustomize_path, "w", encoding="utf-8") as file_ptr:
            yaml.safe_dump(kustomize, file_ptr, sort_keys=False, indent=4)

    def _create_new_tag(self):
        return self._tag_pattern.replace("{{version}}", self._target_version)



import logging
import re
from typing import Optional
from urllib.parse import urlencode

import requests

from ..config.config import LennyBotCheckConfig, LennyBotConfigContainerConfig, LennyBotConfigContainerRegistry
from .icheck import ICheck

PATTERN = r"(?:([\-\_\.\w]+)$)|(?:([\-\_\.\w]+)/([\-\_\.\w]+)$)|(?:([\-\.A-z0-9]+)/([\-\_\.\w]+)/([\-\_\.\w]+)$)|(?:([\-\.A-z0-9]+)/([\-\_\.\w]+)/([\-\_\.\w]+)/([\-\_\.\w]+)$)"


class DockerImage:
    def __init__(self, registry: Optional[str], name: str, tag: Optional[str] = None) -> None:
        self._registry = registry
        self._name = name
        self._tag = tag


class WwwAuthenticateHeader:
    def __init__(self, realm: str, scope: str, service: str) -> None:
        self.realm = realm
        self.scope = scope
        self.service = service

    @classmethod
    def parse(cls, response):
        authenticate = response.headers["Www-Authenticate"]
        realm = authenticate.split('realm="')[1].split('"')[0]
        service = authenticate.split('service="')[1].split('"')[0]
        scope = authenticate.split('scope="')[1].split('"')[0]
        return WwwAuthenticateHeader(realm, scope, service)


class DockerImageAvailableCheck(ICheck):
    def __init__(
        self,
        application_name,
        source_version,
        target_version,
        config: LennyBotCheckConfig,
        container_config: LennyBotConfigContainerConfig,
    ) -> None:
        self._log = logging.getLogger(self.__class__.__name__)
        self._application_name = application_name
        self._source_version = source_version
        self._target_version = target_version
        self._image_pattern = config.image_pattern
        self._container_config = container_config

    @property
    def application(self) -> str:
        return self._application_name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def check(self) -> bool:
        """
        Checks if an image exists in the remote registry from _image_pattern.
        """
        image = self._parse_image()

        if image._registry is None:
            return self._exists_on_docker_hub(image)
        return self._exists_on_registry(image)

    def _parse_image(self):
        """
        Parse function to return a docker image is correct by syntax
        """
        if ":" not in self._image_pattern:
            raise Exception("Image pattern does not contain a tag seperator")

        image_name = self._image_pattern.split(":")[0]
        image_tag = self._image_pattern.split(":")[1].replace("{{version}}", self.target_version)

        match = re.match(PATTERN, image_name)
        if match is None:
            raise Exception(f"Given image pattern is not a valid docker image name {image_name}")
        logging.debug("regex matched following pattern: " + match.group(0))
        if match.group(1) is not None:
            logging.debug("regex matched following pattern: " + match.group(1))
            return DockerImage(None, "library/" + match.group(1), image_tag)
        if match.group(2) is not None:
            logging.debug("regex matched following pattern: " + match.group(2) + "/" + match.group(3) + " " + image_tag)
            return DockerImage(None, match.group(2) + "/" + match.group(3), image_tag)
        if match.group(4) is not None:
            logging.debug(
                "regex matched following pattern: "
                + match.group(4)
                + "/"
                + match.group(5)
                + "/"
                + match.group(6)
                + " "
                + image_tag
            )
            return DockerImage(match.group(4), match.group(5) + "/" + match.group(6), image_tag)
        return DockerImage(match.group(7), match.group(8) + "/" + match.group(9) + "/" + match.group(10), image_tag)

    def _authenticate_on_registry(self, registry: str, authentication_header: WwwAuthenticateHeader) -> str:
        params = {
            "scope": authentication_header.scope,
            "grant_type": "password",
            "service": authentication_header.service,
            "client_id": "lennybot",
            "access_type": "offline",
        }

        url = f"{authentication_header.realm}?{urlencode(params)}"

        if registry in self._container_config.registries.keys():
            registry_data = self._container_config.registries[registry]
            username = registry_data.username
            password = registry_data.password
            if "<REDACTED>" in [username, password]:
                logging.warning(
                    "Either username or password contain '<REDACTED>' and probably have not been overwritten"
                )
            response = requests.get(url, auth=(username, password))
        else:
            logging.debug("Registry not found in config")
            response = requests.get(url)

        if response.status_code == 200:
            token_data = response.json()

            access_token = None
            if "token" in token_data.keys():
                access_token = token_data.get("token")
            elif "access_token" in token_data.keys():
                access_token = token_data.get("access_token")
            else:
                raise Exception("No Access_Token found in response body")

            return str(access_token)

        if response.status_code == 401:
            logging.error("Authentication failed: %d with %s", response.status_code, response.headers)
            raise Exception("Error occurred: Unauthenticated: ", response.status_code)

        if response.status_code == 403:
            logging.error("Authorization failed: %d with %s", response.status_code, response.headers)
            raise Exception("Error occurred: Unauthorization: ", response.status_code)

        if response.status_code == 404:
            logging.error("Nothing Found:  %d with %s", response.status_code, response.headers)
            raise Exception("Error occurred: Nothing Found: ", response.status_code)

        raise Exception("Unexpected Status Code", response.status_code)

    def _exists_on_docker_hub(self, image: DockerImage):
        """
        Checks if the given Docker file exists on DockerHub
        """
        url = f"https://hub.docker.com/v2/repositories/{image._name}/tags?page_size=10000"
        response = requests.get(url)
        if response.status_code != 200:
            raise Exception(f"Unexpected status: {response.status_code} for url: {url}")

        for tag in response.json()["results"]:
            if tag["name"] == image._tag:
                return True
        return False

    def _exists_on_registry(self, image: DockerImage, access_token: Optional[str] = None) -> bool:
        """
        Checks if the given Docker file exists on that perticular registry.
        Also authenticated requests are handled within this function by providing an access token.
        """
        if image._registry is None:
            raise Exception("registry must be set and not be None")

        request_url = f"https://{image._registry}/v2/{image._name}/manifests/{image._tag}"

        # depending on the registry it my helps adding the write accept header :)
        # https://github.com/goharbor/harbor/issues/16075
        headers = {
            "Accept": "application/vnd.oci.image.index.v1+json, application/vnd.docker.distribution.manifest.list.v2+json",
        }
        if access_token is not None:
            headers["Authorization"] = f"Bearer {access_token}"
        response = requests.get(request_url, headers=headers)

        if response.status_code == 401 and access_token is None:
            registry = image._registry
            header_value = WwwAuthenticateHeader.parse(response)
            access_token = self._authenticate_on_registry(registry, header_value)

            return self._exists_on_registry(image, access_token)
        if response.status_code == 200:
            return True
        if response.status_code == 404:
            return False
        raise Exception(f"Unexpected status: {response.status_code} for url {request_url}")



import re

from ...config import LennyBotSourceConfig
from ...helper import semver_2_vc
from ..github import GitHubService
from .isource import ISource


class GithubQuerySource(ISource):
    def __init__(self, name, config: LennyBotSourceConfig, github: GitHubService) -> None:
        self._name = name
        self._github = github
        self._repository = config.repository
        self._version_regex = config.regex

    @property
    def application(self) -> str:
        return self._name

    def latest_version(self):
        tags = self._github.fetch_tags(self._repository)
        results = []
        for tag in tags:
            tag_name = tag["ref"].replace("refs/tags/", "")
            match = re.fullmatch(self._version_regex, tag_name)
            if match is None:
                continue
            if len(match.groups()) < 1:
                continue
            results.append(match.group(1))
        if len(results) == 0:
            raise Exception("No valid version was found")
        results.sort(key=semver_2_vc)
        return results[-1]



import yaml

from ..config.config import LennyBotActionConfig
from .iaction import IAction


class RemoveChecksumsAction(IAction):
    def __init__(self, name, source_version, target_version, config: LennyBotActionConfig) -> None:
        self._name = name
        self._source_version = source_version
        self._target_version = target_version
        if config.target is None:
            raise Exception("Target Path is not set for application " + name)
        self._target_file = config.target

    @property
    def application(self) -> str:
        return self._name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def run(self):
        with open(self._target_file, encoding="utf-8") as file_ptr:
            docs = list(yaml.safe_load_all(file_ptr))

        for doc in docs:
            if doc["kind"] == "Deployment":
                self._update_deployment(doc)
            elif doc["kind"] == "Job":
                self._update_job(doc)

        with open(self._target_file, "w", encoding="utf-8") as file_ptr:
            docs = yaml.safe_dump_all(docs, file_ptr)

    def _update_deployment(self, doc):
        containers = doc["spec"]["template"]["spec"]["containers"]
        for container in containers:
            image = container["image"]
            if "@" not in image:
                continue
            container["image"] = image.split("@")[0]

    def _update_job(self, doc):
        containers = doc["spec"]["template"]["spec"]["containers"]
        for container in containers:
            image = container["image"]
            if "@" not in image:
                continue
            container["image"] = image.split("@")[0]



class ICheck:
    @property
    def application(self) -> str:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass

    @property
    def source_version(self) -> str:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass

    @property
    def target_version(self) -> str:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass

    def check(self) -> bool:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass



import logging
from typing import List

from ..model.plan import LennyBotPlan


class ApplyService:
    def __init__(self) -> None:
        self._log = logging.getLogger(self.__class__.__name__)

    def apply(self, plan: LennyBotPlan):
        state = plan.state
        if not state.is_valid():
            raise Exception("Invalid State")
        for action in plan.actions:
            try:
                action.run()
                state.update_version(action.application, action.target_version)
            except Exception as exception:
                self._log.error(f"Exception during action execution for {action.application}")
                raise exception
        state.save()



import re

import requests

from ...config import LennyBotSourceConfig
from ..github import GitHubService
from .isource import ISource


class GithubSource(ISource):
    def __init__(self, name, config: LennyBotSourceConfig, github: GitHubService) -> None:
        self._name = name
        self._github = github
        self._repository = config.repository
        self._version_regex = config.regex

    @property
    def application(self) -> str:
        return self._name

    def latest_version(self):
        release = self._github.fetch_latest_release(self._repository)
        # TODO check if tag_name property exists
        tag_name = release["tag_name"]
        match = re.fullmatch(self._version_regex, tag_name)
        # TODO check if matched
        if match is None:
            raise Exception(f"Version pattern does not match, Pattern: {self._version_regex}, Tag: {tag_name}")
        if len(match.groups()) < 1:
            raise Exception(f"Missing Group in regex pattern, Pattern: {self._version_regex}, Tag: {tag_name}")
        return match.group(1)



import unittest

from lennybot.check.docker_image_available import DockerImageAvailableCheck, WwwAuthenticateHeader
from lennybot.config.config import LennyBotCheckConfig, LennyBotConfigContainerConfig, LennyBotConfigContainerRegistry


class TestParseImage(unittest.TestCase):
    def setUp(self) -> None:
        self.config = LennyBotCheckConfig()
        self.container_config = LennyBotConfigContainerConfig()

    def test_parse_image_without_tag_separator(self):
        self.config._image_pattern = "missing-colon"
        check = DockerImageAvailableCheck("test-app", "0.2.3", "stable", self.config, self.container_config)
        self.assertRaises(Exception, check.check)

    def test_given_single_image(self):
        self.config._image_pattern = "nginx:stable"
        check = DockerImageAvailableCheck("test-app", "0.2.3", "stable", self.config, self.container_config)
        check.check()

    def test_given_image_with_repo(self):
        self.config._image_pattern = "grafana/grafana:{{version}}"
        check = DockerImageAvailableCheck("test-app", "9.5.5", "9.5.6", self.config, self.container_config)
        check.check()

    def test_given_image_with_registry(self):
        self.config._image_pattern = "quay.io/argoproj/argocd:v{{version}}"
        check = DockerImageAvailableCheck("test-app", "2.7.6", "2.7.7", self.config, self.container_config)
        check.check()

    def test_given_faulty_image_path(self):
        self.config._image_pattern = "qua_y.io/argoproj/argocd:v{{version}}"
        check = DockerImageAvailableCheck("test-app", "2.7.6", "2.7.7", self.config, self.container_config)
        self.assertRaises(Exception, check.check)

    def test_given_image_path_with_4_segments(self):
        self.config._image_pattern = "quay.io/argo/proj/argocd:v{{version}}"
        check = DockerImageAvailableCheck("test-app", "2.7.6", "2.7.7", self.config, self.container_config)
        self.assertRaises(Exception, check.check)

    def test_given_toolong_image_path(self):
        self.config._image_pattern = "quay.io/argo/proj/argo/argocd:v{{version}}"
        check = DockerImageAvailableCheck("test-app", "2.7.6", "2.7.7", self.config, self.container_config)
        self.assertRaises(Exception, check.check)


class TestAuthenticateImage(unittest.TestCase):
    def setUp(self) -> None:
        self.config = LennyBotCheckConfig()
        self.container_config = LennyBotConfigContainerConfig()
        self.docker_image_check = DockerImageAvailableCheck("", "", "", self.config, self.container_config)

    def test_authenticate_on_registry_returns_access_token(self):
        registry = LennyBotConfigContainerRegistry("hub.docker.io")
        header_value = WwwAuthenticateHeader(
            "https://docker-auth.elastic.co/auth", "repository:beats/filebeat:pull", "token-service"
        )
        access_token = self.docker_image_check._authenticate_on_registry(registry.name, header_value)
        self.assertIsNotNone(access_token)

    def test_authenticate_without_credentials_in_config(self):
        self.config._image_pattern = "quay.io/argoproj/argocd:v{{version}}"
        header_value = WwwAuthenticateHeader(
            "https://quay.io/v2/auth", "repository:argoproj/argocd:pull", "token-service"
        )
        access_token = self.docker_image_check._authenticate_on_registry("", header_value)
        self.assertIsNotNone(access_token)

    def test_authenticate_on_registry_wrong_credentials(self):
        self.config._image_pattern = "ghcr.io/brose-ebike/postgres-operator:v{{version}}"
        registry = LennyBotConfigContainerRegistry("ghcr.io")
        registry._username = "ABCD"
        registry._password = "1234"
        self.container_config._registries["ghcr.io"] = registry
        header_value = WwwAuthenticateHeader(
            "https://ghcr.io/v2/auth", "repository:brose-ebike/postgres-operator:pull", "token-service"
        )

        with self.assertRaises(Exception) as context:
            access_token = self.docker_image_check._authenticate_on_registry(registry.name, header_value)
            self.assertIsNone(access_token)
        self.assertFalse("Fails due wrong credentials" in str(context.exception))



import hashlib
import os

import yaml

from ..config.config import LennyBotConfig


class LennyBotState:
    def __init__(self, config: LennyBotConfig) -> None:
        self._filename = config.state_file
        self._init_file()
        self._hash = self._calculate_hash()
        with open(self._filename, encoding="utf-8") as file_ptr:
            self._data = yaml.safe_load(file_ptr)
        if self._data is None:
            self._data = {}

    def _init_file(self):
        if not os.path.exists(self._filename):
            with open(self._filename, "w", encoding="utf-8") as file_ptr:
                yaml.safe_dump({}, file_ptr)

    def _calculate_hash(self):
        with open(self._filename, "rb") as file_ptr:
            return hashlib.sha256(file_ptr.read()).hexdigest()

    def current_version(self, name):
        if name in self._data.keys():
            return self._data[name]["version"]
        return None

    def update_version(self, name, version):
        if name in self._data.keys():
            self._data[name]["version"] = version
        else:
            self._data[name] = {"version": version}

    def is_valid(self):
        return self._hash == self._calculate_hash()

    def save(self):
        if not self.is_valid():
            raise Exception("Invalid State")
        with open(self._filename, "w", encoding="utf-8") as file_ptr:
            yaml.safe_dump(self._data, file_ptr, sort_keys=False, indent=4)



def semver_2_vc(version):
    if version is None:
        return 0
    parts = version.split(".")
    version_code = 0
    for part in parts:
        version_code = version_code * 100 + int(part)
    return version_code



from typing import Any, List

from ..actions.iaction import IAction
from ..model.state import LennyBotState


class LennyBotPlan:
    def __init__(self, state: LennyBotState, actions: List[IAction]) -> None:
        self._state = state
        self._actions = actions

    @property
    def applications(self) -> List[str]:
        result = []
        for action in self._actions:
            result.append(action.application)
        return list(set(result))

    @property
    def actions(self) -> List[IAction]:
        return self._actions

    @property
    def state(self) -> LennyBotState:
        return self._state

    def source_version(self, application: str) -> str | None:
        for action in self._actions:
            if action.application != application:
                continue
            return action.source_version
        return None

    def target_version(self, application: str) -> str | None:
        for action in self._actions:
            if action.application != application:
                continue
            return action.target_version
        return None



import requests

from ..config.config import LennyBotActionConfig
from .iaction import IAction


class DownloadResourcesAction(IAction):
    def __init__(self, name, source_version, target_version, config: LennyBotActionConfig) -> None:
        self._name = name
        self._source_version = source_version
        self._target_version = target_version
        if config.url is None:
            raise Exception("Download URL is not set for application " + name)
        self._url = config.url
        if config.target is None:
            raise Exception("Target Path is not set for application " + name)
        self._target_path = config.target

    @property
    def application(self) -> str:
        return self._name

    @property
    def source_version(self) -> str:
        return self._source_version

    @property
    def target_version(self) -> str:
        return self._target_version

    def run(self):
        download_url = self._url.replace("{{version}}", self._target_version)
        response = requests.get(download_url)
        if response.status_code != 200:
            raise Exception("Unable to download resources")
        with open(self._target_path, "w", encoding="utf-8") as file_ptr:
            file_ptr.write(response.text)



import logging
import os
from typing import Any, Dict, List, Optional

import yaml

CONFIGURATION_OPTIONS = {
    "github": {
        "type": "object",
        "properties": {
            "token": {"type": "string", "required": True, "attribute": "_github_token"},
            "pr": {
                "type": "object",
                "attribute": "_github_pr",
                "properties": {
                    "enabled": {"type": "bool", "attribute": "_enabled"},
                    "repository": {"type": "bool", "attribute": "_repository"},
                    "branchPrefix": {"type": "bool", "attribute": "_branch_prefix"},
                },
            },
        },
    },
    "state": {
        "type": "object",
        "properties": {"file": {"type": "string", "required": True, "attribute": "_state_file"}},
    },
    "logging": {
        "type": "object",
        "properties": {"level": {"type": "string", "required": False, "attribute": "_logging_level"}},
    },
    "container": {
        "type": "object",
        "attribute": "_container",
        "properties": {
            "registries": {
                "type": "map",
                "required": False,
                "attribute": "_registries",
                "value_class": "LennyBotConfigContainerRegistry",
                "properties": {
                    "username": {"type": "string", "attribute": "_username"},
                    "password": {"type": "string", "attribute": "_password"},
                },
            }
        },
    },
    "applications": {
        "type": "list",
        "required": True,
        "class": "LennyBotAppConfig",
        "attribute": "_applications",
        "properties": {
            "name": {"type": "string", "required": True, "attribute": "_name"},
            "source": {
                "type": "object",
                "attribute": "_source",
                "properties": {
                    "type": {"type": "string", "required": True, "attribute": "_type"},
                    "repository": {"type": "string", "required": True, "attribute": "_repository"},
                    "regex": {"type": "string", "attribute": "_regex"},
                },
            },
            "checks": {
                "type": "list",
                "class": "LennyBotCheckConfig",
                "attribute": "_checks",
                "properties": {
                    "type": {"type": "string", "attribute": "_type"},
                    "imagePattern": {"type": "string", "attribute": "_image_pattern"},
                },
            },
            "actions": {
                "type": "list",
                "class": "LennyBotActionConfig",
                "attribute": "_actions",
                "properties": {
                    "type": {"type": "string", "attribute": "_type"},
                    "image": {"type": "string", "attribute": "_image"},
                    "kustomizePath": {"type": "string", "attribute": "_kustomize_path"},
                    "tagPattern": {"type": "string", "attribute": "_tag_pattern"},
                    "url": {"type": "string", "attribute": "_url"},
                    "target": {"type": "string", "attribute": "_target"},
                    "targetFile": {"type": "string", "attribute": "_target_file"},
                    "yamlPath": {"type": "string", "attribute": "_yaml_path"},
                    "valuePattern": {"type": "string", "attribute": "_value_pattern"},
                },
            },
        },
    },
}


class LennyBotSourceConfig:
    def __init__(self) -> None:
        self._type = None
        self._repository = None
        self._regex = None

    @property
    def type(self) -> str:
        return str(self._type)

    @property
    def repository(self) -> str:
        return str(self._repository)

    @property
    def regex(self) -> str:
        return str(self._regex)


class LennyBotCheckConfig:
    def __init__(self) -> None:
        self._type = None
        self._image_pattern: Optional[str] = None

    @property
    def type(self) -> str:
        return str(self._type)

    @property
    def image_pattern(self) -> str:
        return str(self._image_pattern)


class LennyBotActionConfig:
    def __init__(self) -> None:
        self._type = None
        self._image = None
        self._kustomize_path = None
        self._tag_pattern = None
        self._target = None
        self._url = None
        self._target_file = None
        self._yaml_path = None
        self._value_pattern = None

    @property
    def type(self) -> str:
        return str(self._type)

    @property
    def image(self) -> str | None:
        return self._image

    @property
    def kustomize_path(self) -> str | None:
        return self._kustomize_path

    @property
    def tag_pattern(self) -> str | None:
        return self._tag_pattern

    @property
    def target(self) -> str | None:
        return self._target

    @property
    def url(self) -> str | None:
        return self._url

    @property
    def target_file(self) -> str | None:
        return self._target_file

    @property
    def yaml_path(self) -> str | None:
        return self._yaml_path

    @property
    def value_pattern(self) -> str | None:
        return self._value_pattern


class LennyBotAppConfig:
    def __init__(self) -> None:
        self._name = None
        self._source = LennyBotSourceConfig()
        self._checks = []
        self._actions = []

    @property
    def name(self) -> str:
        return str(self._name)

    @property
    def source(self) -> LennyBotSourceConfig:
        return self._source

    @property
    def actions(self) -> List[LennyBotActionConfig]:
        return self._actions


class LennyBotGithubPr:
    def __init__(self) -> None:
        self._enabled = False
        self._repository = None
        self._branch_prefix = "lennybot-"

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def repository(self) -> str:
        return str(self._repository)

    @property
    def branch_prefix(self) -> str:
        return str(self._branch_prefix)


class LennyBotConfigContainerRegistry:
    def __init__(self, name) -> None:
        self._name = name
        self._username: Optional[str] = None
        self._password: Optional[str] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def username(self) -> str:
        return str(self._username)

    @property
    def password(self) -> str:
        return str(self._password)


class LennyBotConfigContainerConfig:
    def __init__(self) -> None:
        self._registries = {}

    @property
    def registries(self) -> Dict[str, LennyBotConfigContainerRegistry]:
        return self._registries


class LennyBotConfig:
    def __init__(self, filename) -> None:
        self._log = None
        with open(filename, encoding="utf-8") as file_ptr:
            self._data = yaml.safe_load(file_ptr)
        self._state_file = None
        self._github_token = None
        self._github_pr = LennyBotGithubPr()
        self._logging_level = "INFO"
        self._container = LennyBotConfigContainerConfig()
        self._applications = []
        self._parse()
        self._configure_logging()

    def _parse(self):
        self._parse_data(CONFIGURATION_OPTIONS, self._data, self)
        self._parse_env()

    def _configure_logging(self):
        logging_level = logging._nameToLevel.get(self._logging_level, logging.DEBUG)
        logging.basicConfig(level=logging_level, format="%(asctime)s [%(levelname)s] %(name)s: %(message)s")
        self._log = logging.getLogger(self.__class__.__name__)
        self._log.debug("Logging was configured")

    def _parse_data(self, schema, data, target):
        for name, property in schema.items():
            if name not in data.keys():
                continue
            config_type = property["type"]
            if config_type in ["object", "list", "map"]:
                self._parse_nested_data(name, property, data, target)
                continue
            attribute_name = property["attribute"]
            property_value = data[name]
            if config_type == "string":
                property_value = str(property_value)
            setattr(target, attribute_name, property_value)

    def _parse_nested_data(self, name, property, data, target):
        attribute_name = None
        config_type = property["type"]
        if "attribute" in property:
            attribute_name = property["attribute"]
            for item in attribute_name.split("."):
                if attribute_name.endswith(item):
                    continue
                target = getattr(target, item)
        if config_type == "object":
            if "attribute" in property:
                target = getattr(target, property["attribute"].split(".")[-1])
            self._parse_data(property["properties"], data[name], target)
            return
        if config_type == "list":
            if attribute_name is None:
                raise Exception("Attribute name was not set, but is needed for config_type list")
            attribute_name = attribute_name.split(".")[-1]
            for item in data[name]:
                array_target = globals()[property["class"]]()
                self._parse_data(property["properties"], item, array_target)
                getattr(target, attribute_name).append(array_target)
            return
        if config_type == "map":
            if "attribute" in property:
                target = getattr(target, property["attribute"].split(".")[-1])
            for key in data[name].keys():
                map_target = globals()[property["value_class"]](key)
                self._parse_data(property["properties"], data[name][key], map_target)
                target[key] = map_target
            return

    def _parse_env(self):
        if "LB_GITHUB_TOKEN" in os.environ.keys():
            self._github_token = os.environ["LB_GITHUB_TOKEN"]
        if "LB_STATE_FILE" in os.environ.keys():
            self._state_file = os.environ["LB_STATE_FILE"]
        for key in os.environ.keys():
            if not key.startswith("LB_CONTAINER_REGISTRY_"):
                continue
            parts = key.removeprefix("LB_CONTAINER_REGISTRY_").split("_")
            registry = parts[0]
            suffix = parts[1]
            registry_data = None
            for name, data in self._container._registries.items():
                if name.lower() == registry.lower():
                    registry_data = data
                    break
            if registry_data is None:
                raise Exception(f"Registry {registry} was not found in configuration")
            if suffix == "USERNAME":
                registry_data._username = os.environ[key]
            elif suffix == "PASSWORD":
                registry_data._password = os.environ[key]
            else:
                raise Exception(f"Unknown suffix: {suffix}")
        # TODO make this generic

    @property
    def state_file(self) -> str:
        return str(self._state_file)

    @property
    def applications(self) -> List[LennyBotAppConfig]:
        return self._applications

    @property
    def github_token(self) -> str | None:
        return self._github_token

    @property
    def github_pr(self) -> LennyBotGithubPr:
        return self._github_pr



from .download_resources import DownloadResourcesAction
from .iaction import IAction
from .remove_checksums import RemoveChecksumsAction
from .update_dockerfile import UpdateDockerfileAction
from .update_image_tag import UpdateImageTagAction
from .update_yaml import UpdateYamlAction


def create_action(name, source_version, latest_version, config) -> IAction:
    action_type = config.type
    if action_type == "image-tag-update":
        return UpdateImageTagAction(name, source_version, latest_version, config)
    if action_type == "download-resources":
        return DownloadResourcesAction(name, source_version, latest_version, config)
    if action_type == "update-yaml":
        return UpdateYamlAction(name, source_version, latest_version, config)
    if action_type == "update-dockerfile":
        return UpdateDockerfileAction(name, source_version, latest_version, config)
    if action_type == "remove-checksums":
        return RemoveChecksumsAction(name, source_version, latest_version, config)
    raise Exception(f"Unknown Action Type: {action_type}")



from typing import Dict, List

import requests
from github import Github
from github.PullRequest import PullRequest
from github.Repository import Repository

from ..config import LennyBotConfig


class GitHubService:
    def __init__(self, config: LennyBotConfig):
        self._config = config
        self._token = self._config.github_token
        self._github = None
        if self._token is not None:
            self._github = Github(self._token)

    def fetch_latest_release(self, repository: str) -> Dict:
        url = f"https://api.github.com/repos/{repository}/releases/latest"
        response = requests.get(url, headers=self._headers())
        if response.status_code != 200:
            raise Exception(f"Unable to fetch latest version, Status: {response.status_code}, Content: {response.text}")
        return response.json()

    def fetch_tags(self, repository: str) -> List:
        url = f"https://api.github.com/repos/{repository}/git/refs/tags"
        response = requests.get(url, headers=self._headers())
        if response.status_code != 200:
            raise Exception(f"Unable to fetch latest version, Status: {response.status_code}, Content: {response.text}")
        return response.json()

    def create_pr(self, branch_name, title, body):
        if self._github is None:
            raise Exception("GitHub is not configured")
        repo = self._github.get_repo(self._config.github_pr.repository)
        new_pull = repo.create_pull(repo.default_branch, branch_name, title=title, body=body)
        labels = self._get_or_create_labels(repo)
        new_pull.add_to_labels(*labels)
        pulls = self._find_own_pulls()
        for pull in pulls:
            if new_pull.id == pull.id:
                continue
            pull.as_issue().create_comment(f"Superseded by #{new_pull.number}")
            pull.edit(state="closed")

    def _find_own_pulls(self) -> List[PullRequest]:
        if self._github is None:
            raise Exception(
                "No github token was given, therefore pull requests can not be listed on "
                + self._config.github_pr.repository
            )
        repo = self._github.get_repo(self._config.github_pr.repository)
        pulls = repo.get_pulls("open")
        result = []
        for pull in pulls:
            if pull.head.ref.startswith(self._config.github_pr.branch_prefix):
                result.append(pull)
        return result

    def _headers(self) -> Dict:
        headers = {}
        if self._token is not None:
            headers["Authorization"] = f"Bearer {self._token}"
        return headers

    def _get_or_create_labels(self, repo: Repository):
        for label in repo.get_labels():
            if label.name == "dependencies":
                return [label]
        label = repo.create_label("dependencies", "0366d6", "Pull requests that update a dependency file")
        return [label]



class ISource:
    @property
    def application(self) -> str:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass

    def latest_version(self) -> str:  # pyright: ignore [reportGeneralTypeIssues, reportReturnType]
        pass



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

import pkg_resources
# Always prefer setuptools over distutils
from setuptools import find_packages, setup


def readme():
    with open("README.md", encoding="utf-8") as f:
        return f.read()


def version():
    with open("version.txt", encoding="utf-8") as f:
        return f.read()


def install_requires():
    with open("requirements.txt", encoding="utf-8") as f:
        reqs = pkg_resources.parse_requirements(f.read())
        return [str(ir) for ir in reqs]


setup(
    name="lennybot",
    version=version(),
    author="Simon Schneider",
    author_email="dev@raynigon.com",
    description="Automatic Updates for Kustomize Resources",
    long_description=readme(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords="kustomize",
    url="http://github.com/raynigon/lennybot",
    project_urls={
        "Bug Reports": "https://github.com/raynigon/lennybot/issues",
        "Source": "https://github.com/raynigon/lennybot/",
    },
    install_requires=install_requires(),
    extras_require={
        "dev": ["setuptools", "wheel", "black"],
        "test": ["coverage", "pytest"],
    },
    python_requires=">=3.6, <4",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    data_files=[("generic", ["version.txt", "README.md", "requirements.txt"])],
    entry_points={
        "console_scripts": [
            "lennybot=lennybot:main",
        ],
    },
)



from lennybot.config.config import LennyBotConfig

from .docker_image_available import DockerImageAvailableCheck
from .icheck import ICheck


def create_check(name, source_version, latest_version, config, global_config: LennyBotConfig) -> ICheck:
    check_type = config.type
    if check_type == "docker-image-available":
        return DockerImageAvailableCheck(name, source_version, latest_version, config, global_config._container)
    raise Exception(f"Unknown Check Type: {check_type}")
